#![deny(clippy::pedantic)]

use scfg::Scfg;
use std::{
    ffi::OsStr,
    fs::OpenOptions,
    io::Read,
    path::PathBuf,
    process::{exit, Stdio},
    str::FromStr,
    time::Duration,
};
use tokio::{
    io::{AsyncBufReadExt, BufReader},
    process::Command,
    try_join,
};

use anyhow::{anyhow, bail, Context};
use domain::{
    base::{Dname, RelativeDname},
    resolv::StubResolver,
};
use lettre::{
    address::Envelope,
    transport::smtp::{
        authentication::{Credentials, DEFAULT_MECHANISMS},
        client::{AsyncSmtpConnection, TlsParameters},
        extension::ClientId,
    },
    Address,
};
use mailparse::{parse_mail, MailAddr, MailHeaderMap, ParsedMail};

#[derive(Debug)]
struct Config {
    inner: Scfg,
}

impl Config {
    fn load() -> anyhow::Result<Config> {
        let config_home = if let Some(s) = std::env::var_os("XDG_CONFIG_HOME") {
            PathBuf::from(s).join("xendmail")
        } else if let Some(home) = std::env::var_os("HOME") {
            PathBuf::from(home).join(".config/xendmail")
        } else if let Ok(cwd) = std::env::current_dir() {
            cwd
        } else {
            bail!("No viable candidate for configuration directory");
        };

        let mut config_file = OpenOptions::new()
            .read(true)
            .write(false)
            .create(false)
            .open(config_home.join("default.scfg"))
            .context("failed to load configuration file")?;

        let mut config = Vec::new();
        config_file.read_to_end(&mut config)?;
        let config = std::str::from_utf8(&config)?;
        let config = config.parse::<Scfg>()?;

        Ok(Config { inner: config })
    }

    fn account(&self) -> anyhow::Result<String> {
        Ok(self
            .inner
            .get("account")
            .context("expected to find directive 'account'")?
            .params()
            .first()
            .context("expected parameter for directive 'account'")?
            .to_string())
    }

    async fn password(&self) -> anyhow::Result<String> {
        let mut parts = self
            .inner
            .get("passcmd")
            .context("Expected to find directive 'passcmd'")?
            .params()
            .iter();
        let cmd = parts
            .next()
            .context("passcmd must contain at least one value")?;
        let mut child = Command::new(cmd)
            .args(parts)
            .stdout(Stdio::piped())
            .spawn()
            .context("Spawn command to retrieve password")?;

        let mut line = String::new();
        BufReader::new(child.stdout.take().expect("child has a piped stdout"))
            .read_line(&mut line)
            .await
            .context("Provided credentials should be valid UTF-8")?;
        let status = child.wait().await.context("Waiting for passcmd to exit")?;
        if !status.success() {
            bail!("Password command exited with status code: {}.", status);
        }
        if line.ends_with('\n') {
            line.pop();
        }
        Ok(line)
    }
}

async fn server_and_port(account: &str) -> anyhow::Result<(String, u16)> {
    let email = Address::from_str(account).context("parsing email address")?;
    let host = email.domain();

    let dname = Dname::bytes_from_str(host)?;
    let candidates = resolve_server(&dname, 587)
        .await
        .context("Failed to resolve SRV records")?;

    let (host, port) = candidates
        .first()
        .context("No results when querying for submission server")?;
    Ok((host.to_string(), *port))
}

/// Open a new connection without authenticating.
///
/// This can be executed while awaiting credentials from the user.
async fn open_connection(account: &str) -> anyhow::Result<AsyncSmtpConnection> {
    let (server, port) = server_and_port(account)
        .await
        .context("Resolve submission server")?;

    let client_id = ClientId::default();
    let mut connection = AsyncSmtpConnection::connect_tokio1(
        (server.as_ref(), port),
        Some(Duration::from_secs(60)),
        &client_id,
        None, // This is for SMTPS, we're using SMTP+STARTTLS (for now).
        None,
    )
    .await
    .context("Connect to resolved submission server")?;
    connection
        .starttls(TlsParameters::new(server)?, &client_id)
        .await
        .context("Requesting STARTTLS upgrade")?;

    Ok(connection)
}

/// Resolves a submission server for a given domain.
///
/// See: <https://www.rfc-editor.org/rfc/rfc6186>
///
/// # Errors
///
/// If the DNS query fails or returns invalid data.
#[allow(clippy::missing_panics_doc)] // Does not really panic.
pub async fn resolve_server<T: std::convert::AsRef<[u8]>>(
    domain: &Dname<T>,
    port: u16,
) -> anyhow::Result<Vec<(String, u16)>> {
    let service = RelativeDname::from_slice(b"\x0b_submission\x04_tcp")
        .expect("submission relative domain is valid");
    let response = StubResolver::new()
        .lookup_srv(service, domain, port)
        .await
        .map_err(|e| anyhow::anyhow!("Error querying SRV records: {:?}", e))?;

    let srvs: Vec<_> = match response {
        Some(s) => s.into_srvs().collect(),
        None => bail!("Submission service decidedly not available"),
    };

    Ok(srvs
        .iter()
        .map(|s| (s.target().to_string(), s.port()))
        .collect())
}

fn extract_from(message: &ParsedMail) -> anyhow::Result<Address> {
    message
        .get_headers()
        .get_first_header("From")
        .map(mailparse::addrparse_header)
        .transpose()?
        .and_then(|addr_list| addr_list.into_inner().into_iter().next())
        .ok_or(anyhow!("Missing 'From' header"))
        .map(|addr| match addr {
            MailAddr::Single(single) => {
                Address::from_str(&single.addr).context("Error parsing address")
            }
            MailAddr::Group(_) => Err(anyhow::anyhow!(
                "The 'From' header cannot contain more than one address"
            )),
        })?
}

fn extract_recipients(message: &ParsedMail) -> Vec<Address> {
    let headers = message.get_headers();
    let recipients = message
        .headers
        .get_all_headers("To")
        .into_iter()
        .chain(headers.get_all_headers("Cc"))
        .chain(headers.get_all_headers("Bcc"))
        .flat_map(mailparse::addrparse_header)
        .flat_map(|addr_list| {
            addr_list
                .into_inner()
                .into_iter()
                .map(|addr| match addr {
                    MailAddr::Group(group) => group.addrs,
                    MailAddr::Single(single) => vec![single],
                })
                .collect::<Vec<_>>()
        })
        .flatten()
        .flat_map(|single| single.addr.parse())
        .collect::<Vec<_>>();

    recipients
}

pub(crate) struct Args {
    /// Recipients of the email message.
    ///
    /// If no recipients are provided, the `To`, `Cc` and `Bcc` fields are used.
    to: Vec<Address>,
}

impl Args {
    fn add_to(&mut self, raw: &OsStr) -> anyhow::Result<()> {
        self.to.push(
            raw.to_str()
                .with_context(|| format!("Email is not valid UTF-8: {}", raw.to_string_lossy()))?
                .parse()?,
        );
        Ok(())
    }

    /// Raw arguments should not include arg0.
    fn parse(raw: impl IntoIterator<Item = impl Into<std::ffi::OsString>>) -> anyhow::Result<Self> {
        let mut args = Args { to: Vec::new() };

        let raw = clap_lex::RawArgs::new(raw);
        let mut cursor = raw.cursor();
        while let Some(arg) = raw.next(&mut cursor) {
            if arg.is_escape() {
                let remaining = raw.remaining(&mut cursor);
                args.to.reserve({
                    let (min, max) = remaining.size_hint();
                    max.unwrap_or(min)
                });
                for r in remaining {
                    args.add_to(r)?;
                }
            } else if arg.is_stdio() {
                // no-op (default is reading from stdin anyway).
            } else if let Some((_long, _value)) = arg.to_long() {
                bail!("Unknown argument {}", arg.display());
            } else if let Some(mut shorts) = arg.to_short() {
                while let Some(short) = shorts.next_flag() {
                    match short {
                        Ok('A' | 'B' | 'b' | 'E' | 'e' | 'i' | 'L' | 'm' | 'o' | 'p' | 'x') => {
                            // No-op (sendmail compat)
                        }
                        Ok(c) => {
                            bail!("Unexpected flag: -{c}");
                        }
                        Err(e) => {
                            bail!("Invalid flag: -{}", e.to_string_lossy());
                        }
                    }
                }
            } else {
                args.add_to(arg.to_value_os())?;
            }
        }

        Ok(args)
    }
}

/// Docs: <https://man.openbsd.org/sendmail>
#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    let mut raw_args = std::env::args_os();
    let prog_name = raw_args
        .next()
        .expect("Argument zero must be the program's name");
    let args = match Args::parse(raw_args) {
        Ok(a) => a,
        Err(err) => {
            eprintln!("{err}");
            eprintln!("Usage: {} to ...", prog_name.to_string_lossy());
            exit(1);
        }
    };
    let mut stdin = std::io::stdin();

    let mut raw = Vec::new();
    stdin.read_to_end(&mut raw)?;
    let parsed = parse_mail(&raw)?;

    let to = if args.to.is_empty() {
        extract_recipients(&parsed)
    } else {
        args.to
    };

    let from = extract_from(&parsed)?;
    let envelope = Envelope::new(Some(from), to)?;
    // TODO: Remove Bcc (need to reconstruct the email for that)

    let config = Config::load().context("Loading configuration file")?;
    let account = config
        .account()
        .context("Reading account from configuration file")?;

    let (password, mut connection) = try_join!(config.password(), open_connection(&account))?;
    connection
        .auth(DEFAULT_MECHANISMS, &Credentials::new(account, password))
        .await?;

    // The `Message` type requires `builder`.
    let response = connection
        .send(&envelope, &raw)
        .await
        .context("Sending message to SMTP server")?;
    if !response.is_positive() {
        eprintln!("Server returned an error response: {:?}", &response);
        exit(1);
    }

    Ok(())
}
